def margin_shallow(entries, h, t):
    return [(k, v) for k, v in entries[:h]] + [(k, v) for k, v in entries[-t:]]


def margin_mapper(entries, kfn, vfn, h, t):
    if vfn is None: vfn = kfn
    return [(kfn(k), vfn(v)) for (k, v) in entries[:h]] + \
           [(kfn(k), vfn(v)) for (k, v) in entries[-t:]]


def margin_mutate(vec, kfn, vfn, h, t):
    if vfn is None: vfn = kfn
    s = (d := len(vec)) - t
    while (h := h - 1) >= 0:
        k, v = vec[h]
        vec[h] = (kfn(k), vfn(v))
    while (d := d - 1) >= s:
        k, v = vec[d]
        vec[d] = (kfn(k), vfn(v))
    return vec



from ject import length


def margin_shallow(vec, h, t):
    return vec[:h] + vec[-t:]


def margin_mapper(vec, fn, h, t):
    if (n := length(fn)) == 1:
        return [fn(x) for x in vec[:h]] + [fn(x) for x in vec[-t:]]
    if n == 2:
        return [fn(x, i) for i, x in enumerate(vec[:h])] + [fn(x, i) for i, x in enumerate(vec[-t:])]
    return margin_shallow(vec, h, t)


def margin_mutate(vec, fn, h, t):
    s = (d := len(vec)) - t
    if (n := length(fn)) == 1:
        while (h := h - 1) >= 0: vec[h] = fn(vec[h])
        while (d := d - 1) >= s: vec[d] = fn(vec[d])
    elif n == 2:
        while (h := h - 1) >= 0: vec[h] = fn(vec[h], h)
        while (d := d - 1) >= s: vec[d] = fn(vec[d], d)
    return vec



def margin_shallow(mx, tp, bt, lf, rt):
    return [row[:lf] + row[-rt:] for row in mx[:tp]] + \
           [row[:lf] + row[-rt:] for row in mx[-bt:]]


def margin_mapper(mx, fn, tp, bt, lf, rt):
    return [[fn(x) for x in row[:lf]] + [fn(x) for x in row[-rt:]] for row in mx[:tp]] + \
           [[fn(x) for x in row[:lf]] + [fn(x) for x in row[-rt:]] for row in mx[-bt:]]


def margin_mutate(mx, fn, tp, bt, lf, rt, ht=None, wd=None):
    s = (ht := ht if ht else len(mx)) - bt
    while (ht := ht - 1) >= s: margin_mutate_row(mx[ht], fn, lf, rt, wd)
    while (tp := tp - 1) >= 0: margin_mutate_row(mx[tp], fn, lf, rt, wd)
    return mx


def margin_mutate_row(row, fn, lf, rt, wd=None):
    s = (wd := wd if wd else len(row)) - rt
    while (wd := wd - 1) >= s: row[wd] = fn(row[wd])
    while (lf := lf - 1) >= 0: row[lf] = fn(row[lf])
    return row



class SparseList(list):
    def __setitem__(self, index, value):
        missing = index - len(self) + 1
        if missing > 0:
            self.extend([None] * missing)
        list.__setitem__(self, index, value)

    def __getitem__(self, index):
        try:
            return list.__getitem__(self, index)
        except IndexError:
            return None



from veho.entries.zipper.zipper import zipper


def duozipper(a, b, kfn, vfn=None):
    return zipper(a, b, kfn, vfn)


def trizipper(a, b, c, kfn, vfn=None):
    if vfn is None: vfn = kfn
    return [(kfn(ak, bk, ck), vfn(av, bv, cv))
            for ((ak, av), (bk, bv), (ck, cv)) in zip(a, b, c)]


def quazipper(a, b, c, d, kfn, vfn=None):
    if vfn is None: vfn = kfn
    return [(kfn(ak, bk, ck, dk), vfn(av, bv, cv, dv))
            for ((ak, av), (bk, bv), (ck, cv), (dk, dv)) in zip(a, b, c, d)]


def to_duozipper(kfn, vfn=None):
    return lambda a, b: duozipper(a, b, kfn, vfn)


def to_trizipper(kfn, vfn=None):
    return lambda a, b, c: trizipper(a, b, c, kfn, vfn)


def to_quazipper(kfn, vfn=None):
    return lambda a, b, c, d: quazipper(a, b, c, d, kfn, vfn)



def margin_shallow(mx, tp, bt, lf, rt):
    return [row[:lf] + row[-rt:] for row in mx[:tp]] + \
           [row[:lf] + row[-rt:] for row in mx[-bt:]]


def margin_mapper(mx, fn, tp, bt, lf, rt):
    return [[fn(x) for x in row[:lf]] + [fn(x) for x in row[-rt:]] for row in mx[:tp]] + \
           [[fn(x) for x in row[:lf]] + [fn(x) for x in row[-rt:]] for row in mx[-bt:]]


def margin_mutate(mx, fn, tp, bt, lf, rt, ht=None, wd=None):
    s = (ht := ht if ht else len(mx)) - bt
    while (ht := ht - 1) >= s: margin_mutate_row(mx[ht], ht, fn, lf, rt, wd)
    while (tp := tp - 1) >= 0: margin_mutate_row(mx[tp], tp, fn, lf, rt, wd)
    return mx


def margin_mutate_row(row, ri, fn, lf, rt, wd=None):
    s = (wd := wd if wd else len(row)) - rt
    while (wd := wd - 1) >= s: row[wd] = fn(row[wd], ri, wd)
    while (lf := lf - 1) >= 0: row[lf] = fn(row[lf], ri, lf)
    return row


# def margin_mutate(mx, fn, tp, bt, lf, rt, ht=None, wd=None):
#     if ht is None:
#         ht = len(mx)
#     s = ht - bt
#     i = 0
#     while i < tp:
#         margin_mutate_row(mx[i], i, fn, lf, rt, wd)
#         i += 1
#     i = s
#     while i < ht:
#         margin_mutate_row(mx[i], i, fn, lf, rt, wd)
#         i += 1
#     return mx


# def margin_mutate_row(row, i, fn, lf, rt, wd=None):
#     if wd is None:
#         wd = len(row)
#     s = wd - rt
#     j = 0
#     while j < lf:
#         row[j] = fn(row[j], i, j)
#         j += 1
#     j = s
#     while j < wd:
#         row[j] = fn(row[j], i, j)
#         j += 1
#     return row



def margin_shallow(vec, h, t):
    return vec[:h] + vec[-t:]


def margin_mapper(vec, fn, h, t):
    return [fn(x) for x in vec[:h]] + [fn(x) for x in vec[-t:]]


def margin_mutate(vec, fn, h, t):
    s = (d := len(vec)) - t
    while (h := h - 1) >= 0: vec[h] = fn(vec[h])
    while (d := d - 1) >= s: vec[d] = fn(vec[d])
    return vec

# if (n := length(fn)) == 1:



from .select import select_values
from .Object import keys, values, entries, to_dict, DICT, SLOTS


class Object:
    @staticmethod
    def keys(ob):
        return list(ob.keys()) if isinstance(ob, dict) \
            else list(ob.__dict__.keys()) if hasattr(ob, DICT) \
            else ob.__slots__ if hasattr(ob, SLOTS) \
            else []

    @staticmethod
    def values(ob):
        return list(ob.values()) if isinstance(ob, dict) \
            else list(ob.__dict__.values()) if hasattr(ob, DICT) \
            else [getattr(ob, s) for s in ob.__slots__] if hasattr(ob, SLOTS) \
            else []

    @staticmethod
    def entries(ob):
        return list(ob.items()) if isinstance(ob, dict) \
            else list(ob.__dict__.items()) if hasattr(ob, DICT) \
            else [(s, getattr(ob, s)) for s in ob.__slots__] if hasattr(ob, SLOTS) \
            else []

    @staticmethod
    def dict(ob):
        return ob if isinstance(ob, dict) \
            else ob.__dict__ if hasattr(ob, DICT) \
            else {s: getattr(ob, s) for s in ob.__slots__} if hasattr(ob, SLOTS) \
            else {}



def init(h, w, func):
    return [[func(i, j) for j in range(w)] for i in range(h)]


def iso(h, w, val):
    return [[val] * w for _ in range(h)]
    # return [[val for _ in range(w)] for _ in range(h)]



def iterate_keys(entries, kfn):
    i, hi = 0, len(entries)
    while i < hi:
        k, = entries[i]
        kfn(k, i)
        i += 1
    pass


def mapper_keys(entries, kfn):
    return [(kfn(k, i), v) for i, (k, v) in enumerate(entries)]


def mutate_keys(entries, kfn):
    i, hi = 0, len(entries)
    while i < hi:
        k, v = entries[i]
        entries[i] = (kfn(k, i), v)
        i += 1
    return entries



def iterate(entries, kfn, vfn=None):
    if vfn is None: vfn = kfn
    i, hi = 0, len(entries)
    while i < hi:
        k, v = entries[i]
        kfn(k, i)
        vfn(v, i)
        i += 1
    pass


def mapper(entries, kfn, vfn=None):
    if vfn is None: vfn = kfn
    return [(kfn(k, i), vfn(v, i)) for i, (k, v) in enumerate(entries)]


def mutate(entries, kfn, vfn=None):
    if vfn is None: vfn = kfn
    i, hi = 0, len(entries)
    while i < hi:
        k, v = entries[i]
        entries[i] = (kfn(k, i), vfn(v, i))
        i += 1
    return entries



from veho.matrix import size


def mutazip(a, b, fn):
    h, w = size(a)
    while (h := h - 1) >= 0:
        ra, rb, y = a[h], b[h], w
        while (y := y - 1) >= 0:
            ra[y] = fn(ra[y], rb[y], h, y)
    return a



from ject import length


def iterate(vec, fn):
    i, hi = 0, len(vec)
    if (n := length(fn)) == 1:
        while i < hi:
            fn(vec[i])
            i += 1
    elif n == 2:
        while i < hi:
            fn(vec[i], i)
            i += 1
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['veho',
 'veho.columns',
 'veho.columns.enumerate',
 'veho.columns.enumerate.mapper',
 'veho.columns.getter',
 'veho.columns.mapper',
 'veho.dict',
 'veho.dict.select',
 'veho.entries',
 'veho.entries.enumerate',
 'veho.entries.enumerate.mapper',
 'veho.entries.enumerate.margin',
 'veho.entries.enumerate.zipper',
 'veho.entries.init',
 'veho.entries.mapper',
 'veho.entries.margin',
 'veho.entries.unwind',
 'veho.entries.zipper',
 'veho.enum',
 'veho.enum.matrix_directions',
 'veho.json',
 'veho.json.init',
 'veho.matrix',
 'veho.matrix.enumerate',
 'veho.matrix.enumerate.mapper',
 'veho.matrix.enumerate.margin',
 'veho.matrix.enumerate.zipper',
 'veho.matrix.init',
 'veho.matrix.mapper',
 'veho.matrix.margin',
 'veho.matrix.select',
 'veho.matrix.utils',
 'veho.matrix.zipper',
 'veho.object',
 'veho.object.Object',
 'veho.object.select',
 'veho.vector',
 'veho.vector.enumerate',
 'veho.vector.enumerate.mapper',
 'veho.vector.enumerate.margin',
 'veho.vector.enumerate.zipper',
 'veho.vector.init',
 'veho.vector.length',
 'veho.vector.mapper',
 'veho.vector.margin',
 'veho.vector.sparse_list',
 'veho.vector.zipper']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'veho',
    'version': '0.0.5',
    'description': 'easy enumerate',
    'long_description': '# veho\n#### easy enumerate\n\n### Usage\n```python\nfrom veho.vector import mapper\nvect = [1,2,3]\nprint(mapper(vect,lambda x:x+1))\n```',
    'author': 'Hoyeung Wong',
    'author_email': 'hoyeungw@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pydget/veho',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .init import init, iso
from .length import length
from .mapper import iterate, mapper, mutate, shallow
from .margin import margin_mapper, margin_mutate, margin_shallow
from .sparse_list import SparseList
from .zipper import duozipper, mutazip, quazipper, to_duozipper, to_quazipper, to_trizipper, trizipper, zipper



from typing import Callable, List

from veho.matrix.utils import size


def iterate(mx: List[list], fn: Callable):
    h, w = size(mx)
    x = 0
    while x < h:
        y = 0
        row = mx[x]
        while y < w:
            fn(row[y])
            y += 1
        x += 1
    pass


def iterate_dev(mx: List[list], fn: Callable):
    for row in mx:
        for cell in row:
            fn(cell)
    pass
